/**
 *  These C routines read data held CDF and CEL files, two of the data formats
 *  used by a well known manufacturer of oligonucleotides arrays.
 *
 *  This is not freeware.
 *  Permission is granted to use this for non-commercial purposes and within the affyR
 *  package.
 *  This was done in the hope it will be usefull, not more. There is no warranty of any kind.
 *
 *
 * Laurent (laurent@cbs.dtu.dk 2001)
 *
 */


#include <R.h>
#include <Rdefines.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef WIN32
#define index strchr
#define rindex strrchr
#endif

/*DEBUG: find a more elegant way to get the size of a character*/
#define SIZE_INT 2
#define SIZE_LINE 160 /* line buffer length... if a line is longer than that... bad, bad, bad... */


/* ************** */
/* #define HAVE_ZLIB */
/* ************** */


#if defined(HAVE_ZLIB)
#include <zlib.h>
#endif


/**
 * ******************EXPORT****************
 */
SEXP readCELfile(SEXP filename, SEXP indexR, SEXP compressR);

SEXP readCDFfile(SEXP filename, SEXP indexR, SEXP compressR);
/**
 * ****************************************
 */


/**
 * Opens a CDF file (making sure it is looks like a CDF file)
 */
int static openCDFfile(const char *filepath, const char *mode, 
		FILE **stream, char *buffy);

/**
 * Opens a CEL file (making sure it is a CEL file)
 */
int static openCELfile(const char *filepath, const char *mode, 
		FILE **stream, char *buffy);


/**
 * A unit is a stuff between '[' and ']'
 * PARAMETERS:
 * -- unit: the name of the unit to go to to
 * -- input_stream: a FILE
 * -- buffy: a character buffer
 * RETURNS:
 * -- 1 found
 * -- 0 not found
 */
int static goToUnit(const char *unit, FILE *input_stream, char *retour);
int static goToUnitStartingWith(const char *unit, FILE *input_stream, char *retour);
char static *getProperty(const char *unit, FILE *input_stream, char *retour);



#if defined(HAVE_ZLIB)
int static gzopenCELfile(const char *filepath, const char *mode, 
		  gzFile *gzstream, char *buffy);

int static gzopenCDFfile(const char *filepath, const char *mode, 
		  gzFile *gzstream, char *buffy);

int static gzgoToUnit(const char *unit, gzFile gzinput_stream, char *retour);
int static gzgoToUnitStartingWith(const char *unit, gzFile gzinput_stream, char *retour);
char static *gzgetProperty(const char *unit, gzFile *gzinput_stream, char *retour);
#endif

/**
 * Fills the data structures
 * PARAMETER(s):
 * -- retour: a string (a line)
 */
void static fillFromLine(char *retour, int nCols, int index, SEXP celfile);

/**
 * Fills the data structures
 * PARAMETER(s):
 * -- retour: a string (a line)
 */
void static fillCharFromLine(char *retour, int nCols, int index, SEXP cdffile, char *buffer);



/*
 * The function called from R 
 * - filename is a character with the path to the file
 * - indexR is an integer to indicate what should be returned
 * - compressR is an integer to indicate whether the file is compressed or not
 */
SEXP readCELfile(SEXP filename, SEXP indexR, SEXP compressR) {
  SEXP celfile,dim;
  FILE *input_stream;
#if defined(HAVE_ZLIB)
  gzFile gzinput_stream;
#endif
  
  char *filepath;
  /*DEBUG*/
  int compress = INTEGER(compressR)[0];
  
  const char *mode = "r";
  const char *param_unit = "HEADER";
  const char *start_unit = "INTENSITY";
  char *retour;
  int numCells;
  int nRows;
  int nCols;
  int tmp;
  int ii;
  int indexVal;
  
  filepath = CHAR(STRING_ELT(filename,0));
  indexVal = INTEGER(indexR)[0];

  retour = R_alloc(SIZE_LINE, SIZE_INT);
  
  if (compress == 0) {
    tmp = openCELfile(filepath, mode, &(input_stream), retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzopenCELfile(filepath, mode, &(gzinput_stream), retour);
#else
    error("Compression not supported on your system !");
#endif
  }
  
  if (tmp == 0) {
    error("The specified file does not appear to be a CEL file.");
  }
  if (tmp == -1) {
    error("Error while trying to open the specified file.");
  }
  if (compress == 0) {
    tmp = goToUnit(param_unit,input_stream,retour);
    /*DEBUG: should test tmp*/
    fgets(retour, SIZE_LINE, input_stream);
    nCols = atoi(retour+5);
    fgets(retour, SIZE_LINE, input_stream);
    nRows = atoi(retour+5);

    tmp = goToUnit(start_unit,input_stream,retour);
    fgets(retour, SIZE_LINE, input_stream);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzgoToUnit(param_unit, gzinput_stream, retour);
    /*DEBUG: should test tmp*/
    gzgets(gzinput_stream, retour, SIZE_LINE);
    nCols = atoi(retour+5);
    gzgets(gzinput_stream, retour, SIZE_LINE);
    nRows = atoi(retour+5);

    tmp = gzgoToUnit(start_unit, gzinput_stream, retour);
    gzgets(gzinput_stream, retour, SIZE_LINE);
#endif
  }

  /*DEBUG: should test this is like 'NumberCells='*/
  numCells = atoi(retour+12);
  
  PROTECT(celfile = NEW_NUMERIC(nRows*nCols));
  PROTECT(dim = NEW_INTEGER(2));
  INTEGER_POINTER(dim)[0] = nRows;
  INTEGER_POINTER(dim)[1] = nCols;
  SET_DIM(celfile,dim);
  
  if (compress == 0) {
    /*NOTE: headers are here...*/
    fgets(retour, SIZE_LINE, input_stream);
    
    for (ii=0; ii<numCells; ii++) {
      fgets(retour, SIZE_LINE, input_stream);
      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CEL file ?).");
	break;
      }
      fillFromLine(retour, nCols, indexVal, celfile);
    }
  } else {
#if defined(HAVE_ZLIB)
    /*NOTE: headers are here...*/
    gzgets(gzinput_stream, retour, SIZE_LINE);
    
    for (ii=0; ii<numCells; ii++) {
      gzgets(gzinput_stream, retour, SIZE_LINE);
      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CEL file ?).");
	break;
      }
      fillFromLine(retour, nCols, indexVal, celfile);
    }
#endif
  }
  UNPROTECT(2);
  return celfile;
}


/** get the 'extra' stuff held in a CEL file (i.e. MASKED, OUTLIERS ...) */
SEXP getIndexExtraFromCEL (SEXP filename, SEXP unitR, SEXP compressR) {
  SEXP extraindex, dim;
  FILE *input_stream;
#if defined(HAVE_ZLIB)
  gzFile gzinput_stream;
#endif

  const char *mode = "r";
  char *filepath;
  int compress;

  char *start_unit;
  char *retour;
  char *buffer;

  int tmp;
  int ii,x,y;
  int numCells;

  filepath = CHAR(STRING_ELT(filename,0));

  start_unit = CHAR(STRING_ELT(unitR,0));

  compress = INTEGER(compressR)[0];

  retour = R_alloc(SIZE_LINE, SIZE_INT);
  buffer = R_alloc(SIZE_LINE, SIZE_INT);

  retour[0] = '\0';
  buffer[0] = '\0';

  if (compress == 0) {
    tmp = openCELfile(filepath, mode, &(input_stream), retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzopenCELfile(filepath, mode, &(gzinput_stream), retour);
#else
    error("Compression not supported onyour system !");
#endif
  }


  if (tmp == 0) {
    error("The specified file does not appear to be a CDF file.");
  }

  if (tmp == -1) {
    error("Error while trying to open the specified file.");
  }


  if (compress == 0) {

    tmp = goToUnit(start_unit, input_stream,retour);
    /*DEBUG: should test tmp*/
    numCells = atoi(getProperty("NumberCells", input_stream, retour));
    /* what's the use ? */
    retour = getProperty("CellHeader", input_stream, retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzgoToUnit(start_unit, gzinput_stream, retour);
    /*DEBUG: should test tmp*/
    numCells = atoi(gzgetProperty("NumberCells", gzinput_stream, retour));
    retour = gzgetProperty("CellHeader", gzinput_stream, retour);
#endif
  }

  PROTECT(extraindex = NEW_NUMERIC(numCells * (2)));
  /* init needed ?*/
  for (ii=0; ii<(numCells * (2)); ii++) {
    NUMERIC_POINTER(extraindex)[ii] = (float)0.0;
  }
  
  PROTECT(dim = NEW_INTEGER(2));
  INTEGER_POINTER(dim)[0] = numCells;
  INTEGER_POINTER(dim)[1] = 2; /* only two columns to hold x and y indexes */
  SET_DIM(extraindex,dim);
 
  if (compress == 0) {

    for (ii=0; ii<numCells; ii++) {

      fgets(retour, SIZE_LINE, input_stream);

      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      if (strlen(retour) <= 1) {
      	fgets(retour, SIZE_LINE, input_stream);
      }

      /* DEBUG */
      /* forced to use 'buffer'... can't figure out why...*/
      
      buffer = retour;
      x = atof(buffer);
      
      buffer = index(buffer,'\t')+1;
      y = atof(buffer);
      
      NUMERIC_POINTER(extraindex)[ii + (numCells) * 0] = x;
      NUMERIC_POINTER(extraindex)[ii + (numCells) * 1] = y;
    }
  } else {
#if defined(HAVE_ZLIB)
    for (ii=0; ii<numCells; ii++) {
      
      gzgets(gzinput_stream, retour, SIZE_LINE);
      
      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      if (strlen(retour) <= 1) {
	gzgets( gzinput_stream, retour, SIZE_LINE);
      }
      
      /* DEBUG */
      /* forced to use 'buffer'... can't figure out why...*/
      buffer = retour;

      x = atof(buffer);
      
      buffer = index(buffer,'\t')+1;
      y = atof(buffer);
      
      NUMERIC_POINTER(extraindex)[ii + (numCells) * 0] = x;
      NUMERIC_POINTER(extraindex)[ii + (numCells) * 1] = y;
    }      
#endif
  }
  UNPROTECT(2);
  return extraindex;  
}

/*
 * The function called from R 
 * - filename is a character with the path to the file
 * - indexR is an integer to indicate what should be returned
 * - compressR is an integer to indicate whether the file is compressed or not
 */

SEXP readCDFfile(SEXP filename, SEXP indexR, SEXP compressR) {
  SEXP cdffile,dim;
  FILE *input_stream;
#if defined(HAVE_ZLIB)
  gzFile gzinput_stream;
#endif
  
  char *filepath;
  /*DEBUG*/
  int compress = INTEGER(compressR)[0];
  
  const char *mode = "r";
  const char *param_unit = "Chip";
  const char *start_unit = "Unit";
  char *retour;
  char *buffer;
  int numUnits;
  int numCells;
  int nRows;
  int nCols;
  int tmp;
  int ii;
  int jj;
  int indexVal;
  
  filepath = CHAR(STRING_ELT(filename,0));
  indexVal = INTEGER(indexR)[0];
  
  retour = R_alloc(SIZE_LINE, SIZE_INT);
  buffer = R_alloc(SIZE_LINE, SIZE_INT);
  
  if (compress == 0) {
    tmp = openCDFfile(filepath, mode, &(input_stream), retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzopenCDFfile(filepath, mode, &(gzinput_stream), retour);
#else
    error("Compression not supported on your system !");
#endif
  }
  
  if (tmp == 0) {
    error("The specified file does not appear to be a CDF file.");
  }
  if (tmp == -1) {
    error("Error while trying to open the specified file.");
  }
  if (compress == 0) {
    tmp = goToUnit(param_unit,input_stream,retour);
    /*DEBUG: should test tmp*/
    nCols = atoi(getProperty("Rows", input_stream, retour));
    nRows = atoi(getProperty("Cols", input_stream, retour));
    numUnits = atoi(getProperty("NumberOfUnits", input_stream, retour));
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzgoToUnit(param_unit, gzinput_stream, retour);
    /*DEBUG: should test tmp*/
    nCols = atoi(gzgetProperty("Rows", gzinput_stream, retour));
    nRows = atoi(gzgetProperty("Cols", gzinput_stream, retour));
    numUnits = atoi(gzgetProperty("NumberOfUnits", gzinput_stream, retour));
#endif
  }
  PROTECT(cdffile = NEW_CHARACTER(nRows*nCols));
  PROTECT(dim = NEW_INTEGER(2));
  INTEGER_POINTER(dim)[0] = nRows;
  INTEGER_POINTER(dim)[1] = nCols;
  SET_DIM(cdffile,dim);

  if (compress == 0) {
    for (ii=0; ii<numUnits; ii++) {
      goToUnitStartingWith(start_unit, input_stream, retour);
      /* WHO conceived this data format ????????? */
      retour = getProperty("NumCells", input_stream, retour);
      retour = getProperty("NumCells", input_stream, retour);

      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;

      }

      numCells = atoi(retour);
      
      retour = getProperty("CellHeader", input_stream, retour);

      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      for (jj=0; jj<numCells; jj++) {
	fgets(retour, SIZE_LINE, input_stream);
	
	/* CDF format madness  */
	if (retour == NULL) {
	  UNPROTECT(2);
	  error("Unexpected and premature end of the given file (truncated CDF file ?).");
	  break;
	}
	if (strlen(retour) <= 1) {
	  fgets(retour, SIZE_LINE, input_stream);
	}
	fillCharFromLine(retour, nCols, indexVal, cdffile, buffer);
      }
    }
  } else {
#if defined(HAVE_ZLIB)
    for (ii=0; ii<numUnits; ii++) {
      gzgoToUnitStartingWith(start_unit, gzinput_stream, retour);
      /* WHO conceived this data format ????????? */
      retour = gzgetProperty("NumCells", gzinput_stream, retour);
      retour = gzgetProperty("NumCells", gzinput_stream, retour);
      
      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      numCells = atoi(retour);
      
      retour = gzgetProperty("CellHeader", gzinput_stream, retour);

      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      for (jj=0; jj<numCells; jj++) {
	retour = gzgets(gzinput_stream, retour, SIZE_LINE);
	
	/* CDF format madness  */
	if (retour == NULL) {
	  UNPROTECT(2);
	  error("Unexpected and premature end of the given file (truncated CDF file ?).");
	  break;
	}
	if (strlen(retour) <= 1) {
	  gzgets(gzinput_stream, retour, SIZE_LINE);
	}
	fillCharFromLine(retour, nCols, indexVal, cdffile, buffer);
      }
    }  
#endif
  }
  UNPROTECT(2);
  return cdffile;
}




int static openCELfile(const char *filepath, const char *mode,
		FILE **stream, char *buffy) {
  int ok;
  char *celfileHEADER;
  FILE *input_stream;
  
  celfileHEADER = "[CEL]";
  ok = 0;
  input_stream = fopen(filepath, mode);

  if (input_stream == NULL) {
    ok = -1;
  } else {
    fgets(buffy, SIZE_LINE, input_stream);
    if (strncmp(celfileHEADER, buffy, 4) == 0) {
      ok = 1;
      rewind(input_stream);
    }
  }
  *(stream) = input_stream; 
  return ok;
}

int static openCDFfile(const char *filepath, const char *mode,
		FILE **stream, char *buffy) {
  int ok;
  char *celfileHEADER;
  FILE *input_stream;
  
  celfileHEADER = "[CDF]";
  ok = 0;
  input_stream = fopen(filepath, mode);

  if (input_stream == NULL) {
    ok = -1;
  } else {
    fgets(buffy, SIZE_LINE, input_stream);
    if (strncmp(celfileHEADER, buffy, 4) == 0) {
      ok = 1;
      rewind(input_stream);
    }
  }
  *(stream) = input_stream; 
  return ok;
}

#if defined(HAVE_ZLIB)

int static gzopenCELfile(const char *filepath, const char *mode,
		gzFile *gzstream, char *buffy) {
  int ok;
  char *celfileHEADER;
  gzFile gzinput_stream;
  
  celfileHEADER = "[CEL]";
  ok = 0;
  gzinput_stream = gzopen(filepath, mode);
  
  if (gzinput_stream == NULL) {
    ok = -1;
  } else {
    gzgets(gzinput_stream, buffy, SIZE_LINE);
    if (strncmp(celfileHEADER, buffy, 4) == 0) {
      ok = 1;
      gzrewind(gzinput_stream);
    }
  }
  *(gzstream) = gzinput_stream; 
  return ok;
}

int static gzopenCDFfile(const char *filepath, const char *mode,
		gzFile *gzstream, char *buffy) {
  int ok;
  char *celfileHEADER;
  gzFile gzinput_stream;
  
  celfileHEADER = "[CDF]";
  ok = 0;
  gzinput_stream = gzopen(filepath, mode);
  
  if (gzinput_stream == NULL) {
    ok = -1;
  } else {
    gzgets(gzinput_stream, buffy, SIZE_LINE);
    if (strncmp(celfileHEADER, buffy, 4) == 0) {
      ok = 1;
      gzrewind(gzinput_stream);
    }
  }
  *(gzstream) = gzinput_stream; 
  return ok;
}
#endif


void static fillFromLine(char *retour, int nCols, int indexVal, SEXP celfile) {
  int x;
  int y;
  double val;
  
  x = atoi(retour);
  retour = index(retour,'\t')+1;
  y = atoi(retour);
  retour = index(retour,'\t')+1;
  if (indexVal == 2) {
    retour = index(retour,'\t')+1;
  }
  val = atof(retour);
  /*SET_ELEMENT(celfile, x + nCols * y, val); */
  NUMERIC_POINTER(celfile)[x + nCols * y] = val;
}



void static fillCharFromLine(char *retour, int nCols, int indexVal, SEXP cdffile, char *buffer) {
  int x;
  int y;
  int ii;
  int stringLen;

  /* Les voies du format sont impenetrables (sic).... */
  retour = index(retour,'=')+1;
  x = atoi(retour);
  retour = index(retour,'\t')+1;
  y = atoi(retour);
  for (ii=0; ii<indexVal; ii++) {
    retour = index(retour,'\t')+1;
  }
  
  stringLen = strcspn(retour, "\t");

  buffer = strncpy(buffer, retour, stringLen);
  /* to end the line...*/
  buffer[stringLen] = '\0';
  
  /*SET_ELEMENT(celfile, x + nCols * y, val); */
  /*NUMERIC_POINTER(celfile)[x + nCols * y] = val;*/
  SET_STRING_ELT(cdffile, x + nCols * y, mkChar(buffer));
    
}


int static goToUnit(const char *unit, FILE *input_stream, char *buffy) {
  int size;
  int found;
  
  char *search_string;
  search_string = R_alloc( strlen(unit)+2, SIZE_INT);
  search_string[0]= '\0';
  strcat(search_string,"[");
  strcat(search_string,unit);
  strcat(search_string,"]");
  size = strlen(search_string);

  found = 0;
  fgets(buffy, SIZE_LINE, input_stream);
  while(buffy != NULL) {
/*      if ((strlen(buffy)-2 == size) */
/*  	&&	(strncmp(search_string, buffy, size) == 0)) { */
    if (strncmp(search_string, buffy, size) == 0) {
      found = 1;
      break;
    }
    fgets(buffy, SIZE_LINE, input_stream);
  } 
  return found;
} 


int static goToUnitStartingWith(const char *unit, FILE *input_stream, char *buffy) {
  int size;
  int found;
  
  char *search_string;
  search_string = R_alloc( strlen(unit)+1, SIZE_INT);
  search_string[0]= '\0';
  strcat(search_string,"[");
  strcat(search_string,unit);
  /* DEBUG: check this really a UNIT !!!!!!!!*/
  /* strcat(search_string,"]"); */
  
  size = strlen(search_string);
  
  found = 0;
  fgets(buffy, SIZE_LINE, input_stream); 
  while(buffy != NULL) {
/*      if ((strlen(buffy)-2 == size) */
/*  	&&	(strncmp(search_string, buffy, size) == 0)) { */
    if (strncmp(search_string, buffy, size) == 0) {
      found = 1;
      break;
    }
    fgets(buffy, SIZE_LINE, input_stream);
  }  
  return found;
} 

char static *getProperty(const char *unit, FILE *input_stream, char *buffy) {
  int size;
  char *result_string;
  char *search_string;
  
  search_string = R_alloc( strlen(unit)+1, SIZE_INT);

  search_string[0]= '\0';
  strcat(search_string,unit);
  strcat(search_string,"=");
  
  size = strlen(search_string);
  
  result_string = R_alloc( SIZE_LINE-size, SIZE_INT );

  fgets(buffy, SIZE_LINE, input_stream); 
  while(buffy != NULL) {
    if (strncmp(search_string, buffy, size) == 0) {
      result_string = strcpy(result_string, buffy+size);
      break;
    }
    fgets(buffy, SIZE_LINE, input_stream);
  }
  return result_string;
}


#if defined(HAVE_ZLIB)
int static gzgoToUnit(const char *unit, gzFile gzinput_stream, char *buffy) {
  int size;
  int found;
  
  char *search_string;
  search_string = R_alloc( strlen(unit)+2, SIZE_INT);
  
  search_string[0]= '\0';
  strcat(search_string,"[");
  strcat(search_string,unit);
  strcat(search_string,"]");
  size = strlen(search_string);
  
  found = 0;
  gzgets(gzinput_stream, buffy, SIZE_LINE);
 
  while(buffy != NULL) {
    /*      if ((strlen(buffy)-2 == size) */
    /*  	&&	(strncmp(search_string, buffy, size) == 0)) { */
    if (strncmp(search_string, buffy, size) == 0) {
      found = 1;
      break;
    }
    gzgets(gzinput_stream, buffy, SIZE_LINE);
  }  
  return found;
} 

int static gzgoToUnitStartingWith(const char *unit, gzFile gzinput_stream, char *buffy) {
  int size;
  int found;
  
  char *search_string;
  search_string = R_alloc( strlen(unit)+2, SIZE_INT);
  
  search_string[0]= '\0';
  strcat(search_string,"[");
  strcat(search_string,unit);
  /* DEBUG: check this really a UNIT !!!!!!!!*/
  /*    strcat(search_string,"]"); */
  size = strlen(search_string);
  
  found = 0;
  gzgets(gzinput_stream, buffy, SIZE_LINE);
 
  while(buffy != NULL) {
/*      if ((strlen(buffy)-2 == size) */
/*  	&&	(strncmp(search_string, buffy, size) == 0)) { */
    if (strncmp(search_string, buffy, size) == 0) {
      found = 1;
      break;
    }
    gzgets(gzinput_stream, buffy, SIZE_LINE);
  }  
  return found;
}
 
char static *gzgetProperty(const char *unit, gzFile *gzstream, char *buffy) {
  int size;
  char *result_string;
  char *search_string;
  
  search_string = R_alloc( strlen(unit)+1, SIZE_INT);

  search_string[0]= '\0';
  strcat(search_string,unit);
  strcat(search_string,"=");
  
  size = strlen(search_string);
  
  result_string = R_alloc( SIZE_LINE-size, SIZE_INT );

  gzgets(gzstream, buffy, SIZE_LINE); 
  while(buffy != NULL) {
    if (strncmp(search_string, buffy, size) == 0) {
      result_string = strcpy(result_string, buffy+size);
      break;
    }
    gzgets(gzstream, buffy, SIZE_LINE);
  }  
  return result_string;
}
#endif


/* DEBUG */

/* 
 * If more than one indexvalR it segfaults... no idea why... it seems to be exactly when
 * buffer2=buffer
 */ 

SEXP readQC(SEXP filename, SEXP startunitR, SEXP indexvalR, SEXP compressR) {
 
  const char *mode = "r";
  const char *param_unit = "Chip";
   
  SEXP qcindex,dim;
  FILE *input_stream;
#if defined(HAVE_ZLIB)
  gzFile gzinput_stream;
#endif
  
  char *filepath;
  int compress;
 
  char *start_unit;
  char *retour;
  char *buffer;
  char *buffer2;
  char *field;
  
  int numVals;
  int numCells;
  int tmp;
  int ii, jj, kk;
  int stringLen;

  int x, y;
  int indexval;

  filepath = CHAR(STRING_ELT(filename,0));
  
  start_unit = CHAR(STRING_ELT(startunitR,0));
  numVals = GET_LENGTH(indexvalR);
  compress = INTEGER(compressR)[0];
  
  retour = R_alloc(SIZE_LINE, SIZE_INT);
  buffer = R_alloc(SIZE_LINE, SIZE_INT);
  buffer2 = R_alloc(SIZE_LINE, SIZE_INT);
  field = R_alloc(SIZE_LINE, SIZE_INT);

  retour[0] = '\0';
  buffer[0] = '\0';
  buffer2[0] = '\0';
  
  if (compress == 0) {
    tmp = openCDFfile(filepath, mode, &(input_stream), retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzopenCDFfile(filepath, mode, &(gzinput_stream), retour);
#else
    error("Compression not supported on your system !");
#endif
  }
  
  if (tmp == 0) {
    error("The specified file does not appear to be a CDF file.");
  }
  if (tmp == -1) {
    error("Error while trying to open the specified file.");
  }
  if (compress == 0) {
    tmp = goToUnit(start_unit, input_stream,retour);
    /*DEBUG: should test tmp*/
    numCells = atoi(getProperty("NumberCells", input_stream, retour));
    retour = getProperty("CellHeader", input_stream, retour);
  } else {
#if defined(HAVE_ZLIB)
    tmp = gzgoToUnit(start_unit, gzinput_stream, retour);
    /*DEBUG: should test tmp*/
    numCells = atoi(gzgetProperty("NumberCells", gzinput_stream, retour));
    retour = gzgetProperty("CellHeader", gzinput_stream, retour);
#endif
  }

  PROTECT(qcindex = NEW_NUMERIC(numCells * (numVals+2)));
  for (ii=0; ii<(numCells * (numVals+2)); ii++) {
    NUMERIC_POINTER(qcindex)[ii] = (float)0.0;
  }
  
  PROTECT(dim = NEW_INTEGER(2));
  INTEGER_POINTER(dim)[0] = numCells;
  INTEGER_POINTER(dim)[1] = numVals+2;
  SET_DIM(qcindex,dim);
 
  if (compress == 0) {

    for (ii=0; ii<numCells; ii++) {

      fgets(retour, SIZE_LINE, input_stream);

      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      if (strlen(retour) <= 1) {
      	fgets(retour, SIZE_LINE, input_stream);
      }

      /* DEBUG */
      /* forced to use 'buffer'... can't figure out why...*/
      buffer = index(retour,'=')+1;
      
      x = atof(buffer);
      
      buffer = index(buffer,'\t')+1;
      y = atof(buffer);
      
      NUMERIC_POINTER(qcindex)[ii + (numCells) * 0] = x;
      NUMERIC_POINTER(qcindex)[ii + (numCells) * 1] = y;
      
      for (jj=0; jj<numVals; jj++) {
	
	buffer2 = buffer;
	indexval = INTEGER(indexvalR)[jj];
	/* printf("---%d-->%s",indexval,buffer); */
	for (kk=0; kk<indexval; kk++) {
	  buffer2 = index(buffer2,'\t')+1;
	  /* printf("-->%s",buffer2); */
	  if (buffer2 == NULL) {
	    error("Invalid index number, Sir...");
	  }
	}
      
	stringLen = strcspn(buffer2, "\t");
	
	field = strncpy(field, buffer2, stringLen);
	/* to end the line...*/
	buffer2[stringLen] = '\0';
	/* printf("%s\n",field); */

	NUMERIC_POINTER(qcindex)[ii + (numCells) * (jj+2)] = atof(field);
	
      }
    }
  } else {
#if defined(HAVE_ZLIB)
    for (ii=0; ii<numCells; ii++) {
      
      gzgets(gzinput_stream, retour, SIZE_LINE);
      
      if (retour == NULL) {
	UNPROTECT(2);
	error("Unexpected and premature end of the given file (truncated CDF file ?).");
	break;
      }

      if (strlen(retour) <= 1) {
	gzgets( gzinput_stream, retour, SIZE_LINE);
      }
      
      /* DEBUG */
      /* forced to use 'buffer'... can't figure out why...*/
      buffer = index(retour,'=')+1;
      
      x = atof(buffer);
      
      buffer = index(buffer,'\t')+1;
      y = atof(buffer);
      
      NUMERIC_POINTER(qcindex)[ii + (numCells) * 0] = x;
      NUMERIC_POINTER(qcindex)[ii + (numCells) * 1] = y;
      
      for (jj=0; jj<numVals; jj++) {
	
	buffer2 = buffer;
	indexval = INTEGER(indexvalR)[jj];

	for (kk=0; kk<indexval; kk++) {
	  buffer2 = index(buffer2,'\t')+1;
	  if (buffer2 == NULL) {
	    error("Invalid index number, Sir...");
	  }
	}
      
	stringLen = strcspn(buffer2, "\t");
      
	field = strncpy(field, buffer2, stringLen);
	/* to end the line...*/
	field[stringLen] = '\0';

	NUMERIC_POINTER(qcindex)[ii + (numCells) * (jj+2)] = atof(field);

      }
    }  
#endif
  }

  UNPROTECT(2);
  return qcindex;  
}
